# -*- coding: utf-8 -*-

# =================================================
# osintgpt
#
# Author: @estebanpdl
#
# File: __init__.py
# Description: Package initialization file.
# =================================================

# define package-level variables and constants
__version__ = '0.0.1'
__name__ = 'osintgpt'
__all__ = []

# describition variables
__author__ = 'Esteban Ponce de Leon'
__doc__ = 'A Python OSINT tool using Large Language Models (LLMs)'
